/*
 * t4graph.h --
 *
 *	Main include file for the T4Graph package.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__T4GRAPH_H__
#define	__T4GRAPH_H__

#ifndef	__cplusplus
#error This file can only be included in C++ files
#endif

/*
 * External dependencies:
 */

#include "tcl.h"
#include "e4graph.h"

/*
 * Ensure correct shared library exporting in WIN32:
 */

#ifndef e4_TGRAPHDLL
    #if defined(_WIN32) && !defined(E4_STATIC)
        #ifdef E4_TGRAPHDLL
            #define e4_TGRAPHDLL __declspec(dllexport)
        #else
            #define e4_TGRAPHDLL __declspec(dllimport)
        #endif
    #else
        #define e4_TGRAPHDLL
    #endif
#endif

/*
 * Forward declarations of the internal representation types, so they
 * can refer to each other.
 */

class e4_TGRAPHDLL T4InternalRep;
class e4_TGRAPHDLL T4Storage;
class e4_TGRAPHDLL T4Node;
class e4_TGRAPHDLL T4Vertex;

/*
 * These functions must be declared extern "C" so their signatures will
 * not be mangled by the C++ compilers. This is depended upon by the Tcl
 * extension loading mechanism.
 */

extern "C" {
extern DLLEXPORT int		Tgraph_RealInit(Tcl_Interp *interp);
extern DLLEXPORT int		Tgraph_Init(Tcl_Interp *interp);
extern DLLEXPORT int		Tgraph_SafeInit(Tcl_Interp *interp);
extern DLLEXPORT int		Tgraphd_Init(Tcl_Interp *interp);
extern DLLEXPORT int		Tgraphd_SafeInit(Tcl_Interp *interp);
}

/*
 * This structure defines options to be passed to e4_Storage;:SetState
 * or an e4_Storage constructor.
 */

typedef struct T4StorageOptions {
    char *driver;		/* String name of storage driver to use. */
    char *rwmode;		/* Read/write mode for storage. */
    int modes;			/* Behavior modes to be set on the newly
				 * opened storage. */
} T4StorageOptions;

/*
 * Enumeration of the kinds of vertex names. Right now, T4Graph supports
 * two kinds of vertex names:
 *
 * T4VNK_INDEX		"n(3)" or simply "n"
 * T4VNK_RANK		"33" or "-1"
 */

typedef enum T4VertexNameKinds {
    T4VNK_INDEX = 0,
    T4VNK_RANK
} T4VertexNameKinds;

/*
 * Declarations of all command procedures:
 */

extern "C" {
extern int		OpenStorageProc(ClientData cd, Tcl_Interp *interp,
					int objc, Tcl_Obj *CONST objv[]);
extern int		IsOpenStorageProc(ClientData cd, Tcl_Interp *interp,
					  int objc, Tcl_Obj *CONST objv[]);
extern int		TransferStorageProc(ClientData cd, Tcl_Interp *interp,
					    int objc, Tcl_Obj *CONST objv[]);
extern int		ParseStorageOptions(Tcl_Interp *interp, int objc,
					    Tcl_Obj *CONST objv[],
					    T4StorageOptions *options);

extern int		StorageCmdProc(ClientData cd, Tcl_Interp *interp,
				       int objc, Tcl_Obj *CONST objv[]);
extern int		NodeCmdProc(ClientData cd, Tcl_Interp *interp,
				    int objc, Tcl_Obj *CONST objv[]);
extern int		VertexCmdProc(ClientData cd, Tcl_Interp *interp,
				      int objc, Tcl_Obj *CONST objv[]);

/*
 * This procedure cleans up at exit.
 */

extern void		T4Graph_ExitProc(ClientData);
}

/*
 * Helper procedures to manage a per-interpreter registry of open
 * storages.
 */

extern T4Storage *	FindRegisteredStorage(Tcl_HashTable *tblPtr,
					      const char *snm,
					      const char *dnm);
extern void		RegisterStorage(Tcl_HashTable *tblPtr,
					T4Storage *s,
					const char *cmdname,
					const char *snm, const char *dnm);
extern void		UnregisterStorage(Tcl_HashTable *tblPtr,
					  T4Storage *s,
					  const char *cmdname,
					  const char *snm,
					  const char *dnm);

/*
 * Create Tcl commands to manipulate a specific T4Graph object:
 */

extern int		MakeStorageCommand(Tcl_Interp *interp, T4Storage *s);
extern char *		MakeNodeCommand(Tcl_Interp *interp, T4Node *np,
					e4_NodeUniqueID nuid);
extern char *		MakeVertexCommand(Tcl_Interp *interp, T4Vertex *vp,
					  e4_VertexUniqueID vuid);

/*
 * Delete Tcl commands that were created to manipulate T4Graph objects:
 */

extern void		DeleteStorageCommand(Tcl_Interp *interp, T4Storage *s);
extern void		DeleteNodeCommand(Tcl_Interp *interp, T4Node *n);
extern void		DeleteVertexCommand(Tcl_Interp *interp, T4Vertex *f);

/*
 * Dispatch to a type-specific object procedure according to the
 * sub-object obtained from the "-with subobj" specification.
 */

extern int		DispatchSubObject(Tcl_Interp *interp,
					  T4Storage *s,
					  T4InternalRep *repPtr,
					  int objc,
					  Tcl_Obj *CONST objv[]);

/*
 * Manage stored procedures and their definition and invocation in Tcl.
 */

extern int		DefineStoredProcedure(Tcl_Interp *interp,
					      char *cmdname,
					      char *nodename,
					      const char *storedproc);
extern int		InvokeStoredProcedure(Tcl_Interp *interp,
					      char *cmdname,
					      int objc,
					      Tcl_Obj *CONST objv[]);

/*
 * Parse a vertex specification. The vertex name can be one of the
 * allowed formats as defined by enum T4VertexNameKinds.
 */

extern int		ParseVertexName(Tcl_Interp *interp, char *spec,
					char **fnp, int *indexp,
					T4VertexNameKinds *vnkp);

/*
 * Parse an insert order name. The name is a string. It is translated
 * by this function into one of the values acceptable for e4_InsertOrder.
 */

extern int		ParseInsertOrder(Tcl_Interp *interp, Tcl_Obj *spec,
					 e4_InsertOrder *iop);

#endif	/* __T4GRAPH_H__ */
